package ActiveState::AR;

use strict;

# This sub is right at the top of the file so that no lexicals are visible to it.
# We do all the work on the stack so that we don't introduce any lexicals of
# our own.
# $_[0] - $self
# $_[1] - $module (e.g. "MODULE")
# $_[2] = a closure defined as
#         sub {
#           # line 1 "$module"     (line directive - man perlsyn)
#           $self->get("lib/$module")
#         }
sub load_module {
    return 1 if $INC{$_[1]};
    die "Can't require $_[1] from $_[0][2]: file not found"
	unless exists $_[0][1]{"lib/$_[1]"};
    $_[2] = scalar eval "sub{\n#line 1 \"$_[1]\"\n" . $_[0]->get("lib/$_[1]") . "}";
    die $@ if $@;
    $_[2]->() or die "$_[1] did not return a true value.\n";
    $INC{$_[1]} = 1;
}

use Compress::Zlib qw(crc32);

our $VERSION = qw(0.01);

# If these look similar to ZIP headers, it's because ActiveState::AR is
# exactly the same as "zip -0" (i.e. no compression).
sub SIG_L() { 0x04034241 } # change 4241 to 4b50 for ZIP compatibility
sub SIG_Z() { 0x04034b50 }
sub HDR_L() { "v3 V4 v2" }

sub SIG()      { "V" }
sub SIG_SIZE() {  4  }
sub HDR_SIZE() { 26  }

# structure:
# [
#   filehandle,
#   { 
#     filename1 => [ filehandle-position,
#                    size,
#                    CRC,
#                  ],
#     filename2...
#   },
#   filename,
# ]

# Read an archive file into an ActiveState::AR object.
#
# Parse filenames from the header immediately. This makes enumeration and
# extraction fast. We don't support compression, which makes life easy.
sub new {
    my ($class, $file) = @_;
    my %files;
    open (my $FH, $file) or die "can't open $file: $!";
    binmode $FH;
    while (1) {
	my $sig = _next_sig($FH) or last;
	next unless $sig == SIG_L
	    # If this line is uncommented, we'll accept uncompressed .zip
	    # files as well as AS::AR files. (But we never emit .zip files.)
	    or $sig == SIG_Z
	;
	my $read = read($FH, my $hdr, HDR_SIZE);
	last unless $read == HDR_SIZE;
	my ($ver,
	    undef, # flags
	    $method,
	    undef, # date
	    $crc,
	    $size_z,
	    $size,
	    $filename_len,
	    $extra_len
	) = unpack(HDR_L, $hdr);
	last if $filename_len == 0;

	# Read the filename
	$read = read($FH, my $filename, $filename_len);
	last unless $read == $filename_len;

	# We require the compression ratio to be 0.
	die "$file: compressed file $filename not supported"
	    if $method != 0 or $size != $size_z;

	last unless seek($FH, $extra_len, 1);
	$files{$filename} = [tell($FH), $size, $crc] if $size;

	# Seek to the end of the content and padding.
	last unless seek($FH, $size, 1);
    }
    bless [$FH, \%files, $file], ref($class) || $class;
}

# Invokes a callback on every file in the archive that matches the regexp.
sub enumerate {
    my ($self, $regexp, $cb, $data) = @_;
    # validate the args
    die "$self->[2]: enumerate requires a regular expression"
	unless defined $regexp and ref($regexp) eq 'Regexp';
    die "$self->[2]: enumerate requires a callback"
	unless defined $cb;
    my $count = 0;
    for my $file (sort keys %{$self->[1]}) {
	next unless $file =~ $regexp;
	++$count;
	&$cb($data, $file, $self->get($file)) or last;
    }
    $count;
}

# Returns the contents of a file from the archive.
sub get {
    my ($self, $file) = @_;
    return unless exists $self->[1]{$file};

    my ($pos, $size, $crc) = @{$self->[1]{$file}};
    seek($self->[0], $pos, 0)
	or die "can't seek to $self->[2]/$file: $!";
    my $read = read($self->[0], my $content, $size);
    die "can't read $self->[2]/$file: $!" unless $read == $size;

    my $calc = crc32($content);
    die "CRC mismatch: $calc != $crc" if $crc ne $calc;

    $content;
}

sub _next_sig {
    my $fh = shift;
    my $read = read($fh, my $buf, SIG_SIZE);
    return 0 unless $read == SIG_SIZE;
    return unpack(SIG, $buf);
}

#############################################################################

package ActiveState::AR::Build;

use Compress::Zlib qw(crc32);

sub VER()   { 20         }
sub SIG_L() { ActiveState::AR::SIG_L() }
sub SIG_G() { 0x02014b50 }
sub SIG_E() { 0x06054b50 }
sub HDR_L() { ActiveState::AR::HDR_L() }
sub HDR_G() { "C2 v3 V4 v5 V2" }
sub HDR_E() { "v4 V2 v" }

# structure:
# [
#   filehandle,
#   file1,
#   file2,
#   ...
# ]

sub new {
    my ($class, $file) = @_;
    open (my $FH, ">", $file) or die "can't write $file: $!";
    binmode $FH;
    bless [$FH], ref($class) || $class;
}

sub dostime {
    my ($s, $m, $h, $d, $M, $y) = localtime;
    $y -= 80;
    $M += 1;

    my $t = 0;
    $t += $s >>  1;
    $t += $m <<  5;
    $t += $h << 11;
    $t += $d << 16;
    $t += $M << 21;
    $t += $y << 25;
    $t;
}

sub add {
    my ($self, $filename, $content) = splice @_, 0, 3;
    my %opts = (    # XXX no explanation for %opts
	hidden => 0,
	@_,
    );

    my $fh = $self->[0];
    my $size = length $content;
    my $pos  = tell $fh;
    my $extra = ''; # per-file comment
    my $t = dostime();
    my $crc = crc32($content);
    print $fh (
	pack(SIG, SIG_L),
	pack(HDR_L,
	    VER,
	    0,  # flags 2
	    0,  # method 2
	    0, # dosDate 4
	    $crc, # 4
	    $size,              # 4
	    $size,              # 4
	    length($filename),  # 2
	    length($extra),     # 2
	),
	$filename,
	$extra,
	$content,
    );
    push @$self, join('', 
	pack(SIG, SIG_G),
	pack(HDR_G,
	    VER,
	    3,  # file attribute format = UNIX 2
	    VER,
	    0,  # flags
	    0,  # method
	    $t, # date 4
	    $crc,   # 4
	    $size,  # 4
	    $size,  # 4
	    length($filename),  # 2
	    length($extra),     # 2
	    0,  # size(comment) 2
	    0,  # disk nm start 2
	    0,  # zipfi_internal 2
	    0,  # zipfi_external 2
	    $pos,   # offset of local file 4
	),
	$filename,
	$extra,
    ) unless $opts{hidden};
}

sub finish {
    my ($self, $comment) = @_;
    $comment = '' unless defined $comment;
    my $fh = shift @$self;
    my $num = @$self;
    my $size = 0;
    my $pos = tell $fh;

    for (@$self) {
	$size += length;
	print $fh $_;
    }

    # Global comment gets tacked on the end
    my $clen = length $comment;
    print $fh (
	pack(SIG, SIG_E),
	pack(HDR_E,
	    0,      # disk number 2
	    0,      # id of last disk 2
	    $num,   # number of entries in central dir 2
	    $num,   # number of entries on this disk 2
	    $size,  # size of central directory 4
	    $pos,   # offset of central directory 4
	    $clen,  # sizeof global comment 2
	),
	$comment,
    );

    close $fh or die $!;
}

#############################################################################

package ActiveState::AR::Overlay;

use Carp;

# structure:
# [
#   { 
#     filename1 => ar-objectn,
#     filename2 => ar-objectn,
#   },
#   ar-object1,
#   ar-object2,
#   ...
# ]

sub new {
    my $class = shift;
    my $self = bless [undef, @_], ref($class) || $class;

    my %idx;
    for my $ar (@$self[1 .. $#$self]) {
	$ar->enumerate(qr/./, sub {
	    my ($data, $file) = @_;
	    my ($ar, $idx) = @$data;
	    $idx->{$file} = $ar;
	}, [$ar, \%idx]);
    }
    $self->[0] = \%idx;

    $self;
}

# Invokes a callback on every file in all archives that matches the regexp
sub enumerate {
    my ($self, $regexp, $cb, $data) = @_;
    die "AS::AR::Overlay: enumerate requires a regular expression"
	unless defined $regexp and ref($regexp) eq 'Regexp';
    die "AS::AR::Overlay: enumerate requires a callback"
	unless defined $cb;
    my $count = 0;
    for my $file (sort keys %{$self->[0]}) {
	my $ar = $self->[0]{$file};
	next unless $file =~ $regexp;
	++$count;
	&$cb($data, $file, $ar->get($file)) or last;
    }
    $count;
}

sub get {
    my ($self, $file) = @_;
    return unless exists $self->[0]{$file};
    return $self->[0]{$file}->get($file);
}

sub load_module {
    my ($self, $mod) = @_;
    croak "Can't require $mod: file not found"
	unless exists $self->[0]{"lib/$mod"};
    return $self->[0]{"lib/$mod"}->load_module($mod);
}

1;
