package ActiveState::Genotype;

use strict;
use ActiveState::DB;
use ActiveState::RangeIndex;
use ActiveState::PCRE;

BEGIN {
    our $VERSION = '0.001';
    require XSLoader;
    XSLoader::load(__PACKAGE__, $VERSION);
}

sub MANAGER() { 0 }
sub BLOB()    { 1 }
sub GENO()    { 2 }
sub RIDX()    { 3 }
sub PCRE()    { 4 }
sub LOGGER()  { 5 }

sub new {
    my ($class, $db) = @_;
    my $manager = ActiveState::DB::Manager->new($db);
    my $self = bless [
	$manager,
	$manager->nest("blob"),
	$manager->nest_object("geno", "ActiveState::Genotype::DB"),
	$manager->nest_object("ridx", "ActiveState::RangeIndex::DB"),
	ActiveState::PCRE->new($manager->nest("pcre")),
	undef,
    ], ref($class) || $class;

    # Find out how many stacks we need at once.
    my $numstacks = $self->[RIDX]->maxvalues;

    # Initialise the genoexec engine.
    $self->[GENO]->init($numstacks);

    $self;
}

sub logger {
    my ($self, $logger) = @_;
    $self->[LOGGER] = $logger;
}

sub pcre_limit {
    my ($self, $limit) = @_;
    $self->[PCRE]->set_limit($limit);
}

sub log {
    my ($self, $level, $msg) = @_;
    return unless $self->[LOGGER];
    $self->[LOGGER]->log($level, $msg);
}

sub match {
    my ($self, $eng, $xtra, $msg) = @_;

    my $bsize = $msg->get_body_size;
    $self->log("DEBUG", "genotype: message has size $bsize");

    # Select the genotypes which could match:
    my @ids = $self->[RIDX]->query($bsize) or return;
    $self->log("DEBUG", "genotype: selected genotypes [@ids] for running.");

    # Prepare the parallel engines for running.
    $self->[GENO]->setup(@ids);

    # The real work happens in part_scan, which is recursively called by
    # $msg->enumerate_parts.
    part_scan([$self, $eng, $xtra, undef, 1], $msg);
}

sub part_scan {
    my ($d, $part) = @_;
    my ($self, $eng, $xtra, $parent, $this) = @$d;
    my $partid = $parent ? "$parent.$this" : $this;

    $self->log("DEBUG", "genotype: scanning part $partid");

    # Run the loaded genotypes against $part. If we no longer need to
    # continue, either because a genotype matched, or they all failed, run()
    # returns false. If it matched, the rule will have been fired by the XS
    # code.
    $self->[GENO]->run(
	$part,
	1,              # shortcircuit, please!
	$eng,           # needed for $eng->addfeature()
	$xtra,          # contains parts, feature hits, and score
	$self->[BLOB]->_asdb,
	$self->[PCRE],
	$partid,        # current part id
    ) or return 0;

    # Recurse further
    $part->enumerate_parts(\&part_scan, [$self, $eng, $xtra, $partid, 1]);

    # Increment the part id
    ++$d->[4];
    return 1;
}

1;

=head1 NAME

ActiveState::Genotype - small and fast "genotype" compiler and runtime

=head1 SYNOPSIS

 # Compile a genotype definition
 use ActiveState::Genotype::Compiler;

 $cc = ActiveState::Genotype::Compiler->new;
 $cc->campaign(...);
 $cc->build("file");

 # Match a message against a set of compiled genotypes:
 use ActiveState::Genotype;
 $matcher = ActiveState::Genotype->new("file");
 $matcher->match($message);

=head1 DESCRIPTION

ActiveState::Genotype is a small and fast "genotype" compiler and runtime. The
runtime consists of a data stack, several registers, and some callback methods
provided by the application to do "hard stuff". A genotype program is a linear
sequence of opcodes with no procedures or branching, although there are
several "stop running if" opcodes.

Genotype is used whenever code needs to run fast and cannot be written in
Perl. Normally Perl code is fast enough, but in some cases Perl uses far too
much memory, or the code needs to be dynamically generated, which means it
cannot be shared across interpreters.

The following sections illustrate how to compile and run "genocode" in several
different ways.

=head2 ActiveState::Genotype::CampaignCompiler

This class knows how to compile C<genotype> definition files into genocode. It
takes a file formatted like this:

 body_min = 100
 body_max = 323

 <match>
     part = b_paras
     min = 3
     max = -3
     test = !/foo{1,7}/
 </match>

and compiles it into an ActiveState::DB database with the following
sub-components:

=over 4

=item o

An ActiveState::RangeIndex index.

Provides fast lookups of definitions by the body size of a message.

=item o

A string table.

Genocode always refers to strings by an index into the string table.

=item o

A PCRE table.

If the C<genotype> definition uses any regular expressions, which most do,
they are stored here. Genocode always refers to regular expressions by an
index into the PCRE table.

=item o

A table of compiled Genocode programs.

=back

The compiler provides the following methods:

=over

=item new()

 $cc = ActiveState::Genotype::CampaignCompiler->new;

Creates a new Genotype Campaign compiler. Takes no arguments.

=item campaign()

 $cc->campaign(%opts);

Adds a new Genotype campaign definition.

The %opts are:

=over

=item class

The C<class> of campaign: C<offensive>, C<health>, etc.

=item id

The C<id> of the campaign. This can be anything you want.

=item rule

The C<rule> to fire if this definition matches. Usually this will be the same
for each campaign in a C<class>, but this is not enforced.

=item source

The raw text of the definition file.

=back

=item finish()

 my $db = $cc->finish;

Returns an C<ActiveState::DB::Build> database, which contains some opaque keys
that you don't need to know about.

=back

=head2 ActiveState::Genotype

This class matches an unclassified message against a compiled set of genotype
definitions.

=over

=item new()

 $m = ActiveState::Genotype->new($db);

Creates a new Genocode matcher object from the C<ActiveState::DB> object
C<$db>.

This class is optimized to match at most a I<single> campaign as quickly as
possible. It computes the maximum stack depth of any program, and the maximum
number of stacks it could possibly need to run all the engines in parallel,
and it allocates enough memory right away.

Whenever you call C<match()> against it, all Genocode programs that could
match the message based on the RangeIndex are executed in parallel until all
fail, or one matches.

=item match()

 $m->match($engine, $xtra, $message);

Match the C<$message> against the compiled genotype definitions. If it
matches, the appropriate rule is fired.

=back

=head2 ActiveState::Genotype::DB

A class indended to be used with C<ActiveState::DB>'s C<nest_object()> call.
It controls one or more genoexec runtime engines, and can be used to run
genocode in parallel or sequentially.

=over

=item new()

Not for public consumption. Create it like so:

 $db = ActiveState::DB->new("file");
 $dbm = ActiveState::DB::Manager->new($db);
 $geno = $dbm->nest_object("KEY", "ActiveState::Genotype::DB");

=item refresh()

Not for public consumption. Called by C<ActiveState::DB::Manager> whenever the
database is refreshed.

=item maxdepth()

 $depth = $geno->maxdepth;

Returns the maximum stack depth required to run any program in the database.

=item init()

 $geno->init($numstacks);

Initialises C<$numstacks> genoexec stacks at once. Each will have depth
C<maxdepth()>, so that any stack can be used to run any genocode.

NOTE: if C<$numstacks> is 0 or omitted, sets up enough stacks to run every
Genocode program in the database at once.

=item reset()

 $geno->reset;

Resets each genoexec stack. Does not modify the code loaded into each stack.

=item setup()

 $geno->setup(@ids);

Resets each genoexec stack, and looks up the compiled program for each of the
C<@ids> in the database. Each program is loaded into the stacks for execution
when C<run()> is called.

NOTE: if setup() is called with no parameters, it sets itself up to run every
id in the database in parallel.

=item run()

 $geno->run($message, $shortcircuit, $engine, $extra, $blobdb, $pcre, $partid);

Runs every genoexec stack in parallel, returning TRUE if the application
should move on to the next mimepart and call run() again, or FALSE if no
further calls to run() are required. Returns when every program has failed, or
all programs are SUSPENDed. Genocode programs are SUSPENDed whenever they need
to descend into a multipart message, because it requires the calling
application to hand control to the message's enumerate_parts() method.

If C<$shortcircuit> is true, returns as soon as any Genocode program SUCCEEDs;
otherwise it runs every program.

The parameters are:

=over

=item $message

The ActiveState::SpamEngine::Message object.

=item $shortcircuit

A boolean; if true, run() returns false as soon as a Genocode program
SUCCEEDs.

=item $engine

The ActiveState::SpamEngine object. Used to fire rules.

=item $extra

The C<$extra> object passed to each ActiveState::SpamEngine callback.

=item $blobdb

An ActiveState::DB object used to look up strings by an integer.

=item $pcre

An ActiveState::PCRE object, used to look up regular expressions by an
integer. May be omitted if you know the programs do not run regular
expressions.

=item $partid

The part id of the C<$message>. May be omitted if you know the programs do not
care.

=back

=item run_seq()

 $geno->run_seq(\@ids, $message, $shortcircuit, $engine, $xtra, $blobdb, $pcre)

Runs every genocode program identified by the C<@ids> in the database. Only a
single stack is required (you must call C<init(1)> before calling run_seq()).
Each program is looked up from the database in turn; ids not found are
skipped.

run_seq() does not support SUSPENDed programs, so if a program SUSPENDs, it is
skipped. That's also why the C<$partid> parameter is missing -- it is supplied
as C<1> to all programs.

If C<$shortcircuit> is true, returns as soon as a Genocode program SUCCEEDs;
otherwise it runs every program.

=head2 ActiveState::Genotype::Utils

ActiveState::Genotype::Utils can export the underlying opcodes for the
runtime:

 use ActiveState::Genotype::Utils qw(:opcodes);

Opcodes are named C<GENOP_PUSH>, C<GENOP_POP>, and so on. You can import only
the ones you need, or all of them with the C<:opcodes> tag.

In addition, ActiveState::Genotype::Utils provides three useful utility
functions:

=over

=item parse_meta_expr()

 ($tokens, @dependencies) = parse_meta_rule("FOO && BAR && BAZ");

Parses a meta rule expression and returns a set of tokens, plus all the
rules referenced by the expression.

=item compile_meta_tokens()

 @ops = compile_meta_tokens($tokens);

Turns the C<$tokens> returned by C<parse_meta_rule()> into a list of Genocode
opcodes. The opcodes are not yet in compiled form, so that they can be
manipulated before being compiled.

See L</ActiveState::Genotype::CompilerXS> for how to compile them.

=item compile_meta_expr()

 @ops = compile_meta_expr("FOO && BAR && BAZ");

Calls compile_meta_tokens() on the output of parse_meta_expr().

=back

=head2 ActiveState::Genotype::CompilerXS

This class should not be C<use>d, because it springs into existence when you
C<use> ActiveState::Genotype. It provides low-level access to building
Genocode opcode-by-opcode.

=over

=item new()

 $cc = ActiveState::Genotype::CompilerXS->new;

Create an empty program.

=item pushop()

 $cc->pushop(GENOP_PUSH, 22);
 $cc->pushop(GENOP_PUSH, 20);
 $cc->pushop(GENOP_ADD);

Push an opcode onto the end of the program. Optionally adds an argument.

=item compile()

 $string = $cc->compile;

Compiles the opcodes into a binary program. Does a few simple sanity checks
like making sure the stack doesn't underflow before compiling. Croaks on
errors.

=back
