package ActiveState::Genotype::CampaignCompiler;

use strict;
use ActiveState::Config::Simple qw(config_parse);
use ActiveState::DB;
use ActiveState::RangeIndex;
use ActiveState::PCRE qw(qr2pcre);
use ActiveState::Genotype::Utils qw(:opcodes
				    :methods
				    compile_meta_tokens parse_meta_expr);

sub new {
    my ($cls) = @_;
    bless {
	blob    => {},
	pcre    => ActiveState::PCRE::Build->new,
	code    => [],
	ridx    => ActiveState::RangeIndex->new,
	keep    => 1,
	part    => {},
    }, ref($cls) || $cls;
}

sub valid_parts {
    my ($self, $parts) = @_;
    $self->{part} = $parts;
}

sub keep_source {
    my ($self, $keep) = @_;
    $self->{keep} = $keep;
}

sub _blob_id {
    my ($self, $string) = @_;
    return $self->{blob}{$string} if defined $self->{blob}{$string};
    return $self->{blob}{$string} = ++$self->{__blobid__};
}

sub _pcre_id {
    my ($self, $regexp) = @_;
    return $self->{pcre}->compile($regexp);
}

sub _code_id {
    my ($self, $prog) = @_;
    push @{$self->{code}}, $prog;
    return $#{$self->{code}};
}

sub campaign {
    my $self = shift;
    my %args = @_;

    my $src    = delete $args{source} || die "argument 'source' required";
    my $class  = delete $args{class}  || die "argument 'class' required";
    my $id     = delete $args{id}     || die "argument 'id' required";
    my $rule   = delete $args{rule}   || die "argument 'rule' required";
    my $refcnt = delete $args{refcnt};
    die "Unknown option '$_'" for keys %args;

    my $d = eval { config_parse($src) };
    die "error in $class/$id: $@" if $@;

    my ($min, $max) = delete @$d{qw(body_min body_max)};
    unless (defined $min) {
	die "error in $class/$id: no body_min defined";
    }
    unless (defined $max) {
	die "error in $class/$id: no body_max defined";
    }

    my $prog = ActiveState::Genotype::CompilerXS->new;

    if ($d->{meta}) {
	my ($tokens, @deps) = parse_meta_expr(delete $d->{meta});
	if ($refcnt) {
	    ++$refcnt->{$_} for @deps;
	}
	my @expr = compile_meta_tokens($tokens);
	for my $term (@expr) {
	    if (ref($term) eq 'ARRAY') {
		my ($op, $arg, $is_string) = @$term;
		$arg = $self->_blob_id($arg) if $is_string;
		$prog->pushop($op, $arg);
	    }
	    else {
		$prog->pushop($term);
	    }
	}
	$prog->pushop(GENOP_SKIPNIF, 1);
	$prog->pushop(GENOP_FAIL);
    }
    eval { $self->compile_mimepart("1", $prog, $d) };
    die "error in $class/$id: $@" if $@;

    if ($self->{keep}) {
	$prog->pushop(GENOP_PUSH, $self->_blob_id($rule));
	$prog->pushop(GENOP_PUSH, $self->_blob_id("$class/$id"));
	$prog->pushop(GENOP_ADDFEATURE2);
    }
    else {
	$prog->pushop(GENOP_ADDFEATURE, $self->_blob_id($rule));
    }
    $prog->pushop(GENOP_SUCCEED);

    # Emit the compiled code!
    my $progid = $self->_code_id($prog->compile);
    $self->{ridx}->addrange($min, $max, $progid);
}

sub compile_mimepart {
    my ($self, $partid, $prog, $d) = @_;
    if ($partid ne '1') {
	$prog->pushop(GENOP_MIMEPART, $self->_blob_id($partid));
    }
    if ($d->{body_min}) {
	$prog->pushop(GENOP_BODYSIZE);
	$prog->pushop(GENOP_PUSH, delete $d->{body_min});
	$prog->pushop(GENOP_LT);
	$prog->pushop(GENOP_SKIPNIFZ, 1);
	$prog->pushop(GENOP_FAIL);
    }
    if ($d->{body_max}) {
	$prog->pushop(GENOP_BODYSIZE);
	$prog->pushop(GENOP_PUSH, delete $d->{body_max});
	$prog->pushop(GENOP_GT);
	$prog->pushop(GENOP_SKIPNIFZ, 1);
	$prog->pushop(GENOP_FAIL);
    }

    # The "ct" test is interesting. Unlike <match> blocks, it is not
    # an automatic failure for the message to have no Content-Type header; if
    # $d->{ct} is a substring of the default value, then it's okay.
    if ($d->{ct}) {
	my $nullok = qq{text/plain; charset="us-ascii"} =~ /\Q$d->{ct}\E/;
	$prog->pushop(GENOP_PARTHEADER, $self->_blob_id("content-type"));

	if ($nullok) {
	    $prog->pushop(GENOP_SKIPNIFZ, 3); # ops to skip ==> ...
	}
	else {
	    $prog->pushop(GENOP_SKIPNIF, 1);
	    $prog->pushop(GENOP_FAIL);
	}

	# ... ==> here are the ops.
	$prog->pushop(GENOP_MATCHSTR, $self->_blob_id(delete $d->{ct}));
	$prog->pushop(GENOP_SKIPNIF, 1);
	$prog->pushop(GENOP_FAIL);
    }

    # Ensure that there is a <match> field, even if it's empty.
    my $m = delete $d->{match} || [];
    $m = [$m] unless ref($m) eq 'ARRAY';

    # Turn 'te' into an equivalent <match> block. This allows it to be merged
    # with any other tests on the Content-Transfer-Encoding header.
    if ($d->{te}) {
	push @$m, {
	    part => 'h_content_transfer_encoding',
	    test => delete $d->{te},
	};
    }

    # Combine matches that operate on the same part and range, to avoid
    # setting the part multiple times.
    my $mm = {}; # mm = merged matches
    for my $match (@$m) {
	my ($part, $min, $max, $test) = delete @$match{qw(part min max test)};

	die "match section missing 'part'" unless defined $part;
	die "match section missing 'test'" unless defined $test;
	die "unknown token '$_' in <match>" for keys %$match;
	$min =  0 unless defined $min;
	$max = -1 unless defined $max;

	$test = [$test] if ref $test ne "ARRAY";
	push @{$mm->{$part}{"$min $max"}}, $test;
    }

    for my $part (sort keys %$mm) {
	my $set_range;

	# There are three basic types of part: cached parts from the part
	# plugin; h_* means to operate on a message header; and 
	if ($self->{part}{$part}) {
	    $prog->pushop(GENOP_PARTCACHED, $self->_blob_id($part));
	    $set_range = 1;
	}
	elsif ($part =~ /\Ah_([a-z][a-z0-9_]*)\z/) {
	    (my $header = $1) =~ s/_/-/g;
	    $prog->pushop(GENOP_PARTHEADER, $self->_blob_id($header));
	}
	elsif ($part =~ /^&(.*)/) {
	    $prog->pushop(GENOP_PARTMETHOD, _method_id($1));
	}
	else {
	    die "invalid part $part";
	}
	$prog->pushop(GENOP_SKIPNIF, 1);
	$prog->pushop(GENOP_FAIL);

	for my $range (sort keys %{$mm->{$part}}) {
	    if ($set_range) {
		my ($min, $max) = split ' ', $range;
		$prog->pushop(GENOP_PUSH, $min);
		$prog->pushop(GENOP_PUSH, $max);
		$prog->pushop(GENOP_SETRANGE);
		$prog->pushop(GENOP_SKIPNIF, 1);
		$prog->pushop(GENOP_FAIL);
	    }

	    for my $test (@{$mm->{$part}{$range}}) {
		# Repeated tests always combine with OR. This means we want to
		# bail out of the match block on the first successful test. If
		# no test succeeds, we fail.
		my $INSTRUCTIONS_PER_TEST = 2;
		my $skip = $INSTRUCTIONS_PER_TEST * @$test;
		for my $t (@$test) {
		    my $neg = $self->_emit_test($prog, $t);
		    my $op  = $neg ? GENOP_SKIPNIFZ : GENOP_SKIPNIF;
		    $prog->pushop($op, $skip - 1);
		    $skip -= $INSTRUCTIONS_PER_TEST;
		}
		$prog->pushop(GENOP_FAIL);
	    }
	}
    }

    # Recurse into <mimepart> sections.
    if ($d->{mimepart}) {
	my $mp = delete $d->{mimepart};
	for my $id (sort { $a <=> $b } keys %$mp) {
	    $self->compile_mimepart("$partid.$id", $prog, $mp->{$id});
	}
    }

    die "unknown token '$_'.\n" for keys %$d;
}

my %method_id = (
    get_body            => METHOD_BODY,
    get_decoded_body    => METHOD_DECODED_BODY,
);
sub _method_id {
    my $method = shift;
    die "invalid method \&$method" unless exists $method_id{$method};
    $method_id{$method};
}

sub _emit_test {
    my ($self, $prog, $t) = @_;
    my $negate = $t =~ s/^\s*!\s*//;
    die "'!' character must be followed by a test to negate" unless length $t;
    if (defined (my $pcre = qr2pcre($t))) {
	$prog->pushop(GENOP_MATCHRE, $self->_pcre_id($pcre));
    }
    else {
	$prog->pushop(GENOP_MATCHSTR, $self->_blob_id($t));
    }
    return $negate;
}

sub finish {
    my $self = shift;

    # The blob table
    my $d = $self->{blob};
    my $blob = ActiveState::DB::Build->new;
    $blob->set($d->{$_} => $_) for sort { $d->{$a} <=> $d->{$b} } keys %$d;

    # The program table
    my $comp = ActiveState::DB::Build->new;
    for my $i (0 .. $#{$self->{code}}) {
	$comp->set($i => $self->{code}[$i]);
    }

    my $db = ActiveState::DB::Build->new;
    $db->set(geno => $comp->build);
    $db->set(blob => $blob->build);
    $db->set(pcre => $self->{pcre}->build);
    $db->set(ridx => $self->{ridx}->pack);
    $db;
}

1;
