package ActiveState::Genotype::Utils;

use strict;
use base 'Exporter';

our (@EXPORT_OK, %EXPORT_TAGS);

# XS code injects constants into our namespace, and fills up @EXPORT_OK
use ActiveState::Genotype;

push @EXPORT_OK, qw(parse_meta_expr compile_meta_expr compile_meta_tokens);
%EXPORT_TAGS = (
    all     => [@EXPORT_OK],
    opcodes => [grep /^GENOP/, @EXPORT_OK],
    methods => [grep /^METHOD/, @EXPORT_OK],
    meta    => [qw(parse_meta_expr compile_meta_expr compile_meta_tokens)],
);

#=============================================================================
# Compile meta expressions into genocode snippets.
#=============================================================================
my $rule_re = qr/[A-Z_][A-Z0-9_]*/;
my $int_re  = qr/[0-9]+/;

sub parse_meta_expr {
    my $expr = shift;
    my @tokens = $expr =~ /$rule_re|$int_re|\|\||\&\&|[+><!()]/g;
    my @deps   = grep /^$rule_re$/, @tokens;
    return wantarray ? (\@tokens, @deps) : \@tokens;
}

sub compile_meta_expr {
    my $expr = shift;
    my $tokens = parse_meta_expr($expr);
    return compile_meta_tokens($tokens);
}

my %OP = (
#   OP       CODE       PRECEDENCE
    '||' => [GENOP_OR,  0],
    '&&' => [GENOP_AND, 1],
    '>'  => [GENOP_GT,  2],
    '<'  => [GENOP_LT,  2],
    '+'  => [GENOP_ADD, 3],
    '!'  => [GENOP_NOT, 4],
);

sub compile_meta_tokens {
    my ($expr, $depth) = @_;
    $depth ||= 0;
    my @opstack;
    my @postfix;
    while (@$expr) {
	my $cur = shift @$expr;
	if ($OP{$cur}) {
	    my ($op, $prec) = @{$OP{$cur}};
	    if (@opstack) {
		# while the precedence of $cur is less than the top of the
		# operand stack, emit the operand stack
		while (@opstack) {
		    my $op2 = $opstack[-1];
		    my $prec2 = $OP{$op2}[1];
		    last if $prec2 < $prec;
		    push @postfix, $OP{pop @opstack}[0];
		}
	    }
	    push @opstack, $cur;
	}
	elsif ($cur eq '(') {
	    push @postfix, compile_meta_tokens($expr, $depth + 1);
	    my $t = shift @$expr;
	    die "Mismatched '(' and ')': $t" if $t ne ")";
	}
	elsif ($cur eq ')') {
	    die "Unexpected ')'" unless $depth;
	    unshift @$expr, $cur; # putback
	    last;
	}
	elsif ($cur =~ /^$rule_re$/) {
	    push @postfix, [GENOP_HASFEATURE, $cur, 1];
	}
	else {
	    push @postfix, [GENOP_PUSH, $cur];
	}
    }
    push @postfix, map $OP{$_}[0], reverse @opstack;
    @postfix;
}

1;
