package ActiveState::PCRE;

use strict;
use ActiveState::DB;
use base 'Exporter';

our $VERSION = '1.000';
our @EXPORT_OK = qw(@START @END $LIMIT qr2pcre);
our (@START, @END, $LIMIT);

use XSLoader ();
XSLoader::load(__PACKAGE__, $VERSION);

sub DB()    { 0 }
sub COUNT() { 1 }
sub LIMIT() { 2 }

# Turn a string like "/blah/abcd" into "(?abcd)blah".
sub qr2pcre {
    my $re = shift;
    $re =~ s/\A\s*//;
    $re =~ s/\s*\z//;
    if ($re =~ s#\A/(.*)/(\w*)\z#$1#) {
	$re = "(?$2)$re" if $2;
	return $re;
    }
    else {
	return undef;
    }
}

sub new {
    my ($class, $thing) = @_;

    my $db;
    if (UNIVERSAL::isa($thing, "ActiveState::DB")) {
	$db = $thing;
    }
    else {
	$db = ActiveState::DB->new($thing);
    }

    my $self = bless [$db, $db->get("#PCRE#")], ref($class) || $class;
    die "pcre: malformed database has no PCRE key" unless defined $self->[COUNT];
    $self->[LIMIT] ||= 0;
    $self;
}

# At some point, we could expose these, but for now...
#my %exec_opts = (
#    anchored        => ActiveState::PCRE::PCRE_ANCHORED(),
#    notbol          => ActiveState::PCRE::PCRE_NOTBOL(),
#    noteol          => ActiveState::PCRE::PCRE_NOTEOL(),
#    notempty        => ActiveState::PCRE::PCRE_NOTEMPTY(),
#);

sub set_limit {
    my ($self, $limit) = @_;
    $self->[LIMIT] = $limit;
}

sub match {
    my $self = shift;
    my $id   = shift;
    die "pcre '$id' too big (>= $self->[COUNT])" if $id >=$self->[COUNT];
    my $pcre  = $self->[DB]->get("${id}r") or die "pcre '$id' does not exist";
    my $study = $self->[DB]->get("${id}s");
    my $offset = 0;
    $offset = pop if @_ > 1;
    return pcre_exec($_[0], $pcre, $study, $offset);
}

# when: 1 to call cb on a match; 0 to call cb when there are no matches
# code: Callback should return 0 if one match is enough, or 1 to try to match again

sub match_asdb {
    my ($self, $cb, $data, $asdb, $when, $save) = splice @_, 0, 6;
    $when = 1 unless defined $when;
    $save = 0 unless defined $save;
    my %args = (
	asdb => $asdb,
	code => $cb,
	data => $data,
	when => $when,
	save => $save,
    );
    return match_asdb_v2($self, \%args, @_);
}

sub match_asdb_v2 {
    my ($self, $args) = (shift, shift);
    die 'Usage: match_asdb_v2($self, \%args, @strings)'
	unless $args and ref($args) eq 'HASH';
    _xs_match_asdb(

	# This is the PCRE database. Its keys are compiled regexp ids, and the
	# values are compiled PCREs.
	$self->[DB]->_asdb,

	# This database is iterated over. For each iteration, the key should
	# be a compiled regexp ID used to run the regular expression. If the
	# regexp matches, the value is passed to the callback.
	$args->{asdb}->_asdb,

	# Callback and data
	$args->{code},
	$args->{data},

	# Options
	$args->{logger},
	$args->{when} || 0,
	$args->{save} || 0,
	$args->{limit} || $self->[LIMIT],

	# Strings to match against.
	@_
    );
}

sub count {
    my $self = shift;
    return $self->[COUNT];
}

sub _xs_pcre {
    my $self = shift;
    return _xs_pcre_new($self->[DB]->_asdb, $self->[LIMIT]);
}

package ActiveState::PCRE::Build;

sub new {
    my $class = shift;
    bless {
	id => 0,
	re => {},
	pcre => [],
	study => [],
    }, ref($class) || $class;
}

my %cmpl_opts = (
    caseless        => ActiveState::PCRE::PCRE_CASELESS(),
    multiline       => ActiveState::PCRE::PCRE_MULTILINE(),
    dotall          => ActiveState::PCRE::PCRE_DOTALL(),
    extended        => ActiveState::PCRE::PCRE_EXTENDED(),
    dollar_endonly  => ActiveState::PCRE::PCRE_DOLLAR_ENDONLY(),
    extra           => ActiveState::PCRE::PCRE_EXTRA(),
    ungreedy        => ActiveState::PCRE::PCRE_UNGREEDY(),
    utf8            => ActiveState::PCRE::PCRE_UTF8(),
    no_auto_capture => ActiveState::PCRE::PCRE_NO_AUTO_CAPTURE(),
);

sub compile {
    my $self = shift;
    my $re   = shift;

    my $flags = 0;
    while (@_) {
	my ($opt, $val) = splice(@_, 0, 2);
	die "Unknown PCRE option '$opt'" unless exists $cmpl_opts{$opt};
	if ($val) {
	    $flags |= $cmpl_opts{$opt};
	}
	else {
	    $flags &= ~$cmpl_opts{$opt};
	}
    }
    my $key = "$flags:$re";
    return $self->{re}{$key} if exists $self->{re}{$key};

    my ($pcre, $study) = ActiveState::PCRE::pcre_compile($re, $flags);
    my $id = $self->{id}++;
    $self->{pcre}[$id] = $pcre;
    $self->{study}[$id] = $study;
    return $self->{re}{$key} = $id;
}

sub _build {
    my $self = shift;
    my $db = ActiveState::DB::Build->new;
    my $max = $self->{id};
    $db->set("#PCRE#" => $max);
    for my $id (0 .. $max - 1) {
	$db->set("${id}r" => $self->{pcre}[$id]);
	$db->set("${id}s" => $self->{study}[$id]) if $self->{study}[$id];
    }
    $db;
}

sub build {
    my $self = shift;
    my $db = $self->_build;
    $db->build;
}

sub build_to {
    my $self = shift;
    my $file = shift;
    my $db = $self->_build;
    $db->build_to($file);
}

1;

=head1 NAME

ActiveState::PCRE - use perl-compatible regular expressions from Perl!

=head1 SYNOPSIS

 $compile = ActiveState::PCRE::Build->new;
 $id = $compile->compile('abc|def');
 $compile->build("pcre.asdb");

 $run = ActiveState::PCRE->new("pcre.asdb");
 if ($run->match($id, "def")) {
    print "Matched!\n";
 }

=head1 DESCRIPTION

ActiveState::PCRE is an interface to the C<pcre> library. It loads precompiled
regular expressions from an ActiveState::DB database. Regular expressions can
then evaluated against Perl strings such that the compiled regular expressions
are shared between threads (and processes).

=head2 Compiling Regular Expressions: ActiveState::PCRE::Build

ActiveState::PCRE::Build exposes the following methods, which can be used to
build a read-only database of regular expressions.

=over 

=item new()

 $pcre = ActiveState::PCRE::Build->new;

Create a new ActiveState::PCRE::Build object. Initially it contains no regular
expressions.

=item compile()

 $id = $pcre->compile($regexp, %opts);

Compiles the given C<$regexp> and returns a numeric C<$id> which can be passed
to C<ActiveState::PCRE::match()>.

If C<$regexp> has already been seen, the same C<$id> is returned.

If the regular expression does not compile, C<compile()> dies with an error.

=item build_to()

 $pcre->build_to($filename);

Writes the compiled regular expressions to an ActiveState::DB database called
C<$filename>.

=back

=head2 Matching Strings: ActiveState::PCRE

ActiveState::PCRE provides the following methods which can be used to match a
Perl string against a precompiled regular expression.

=over

=item new()

 $pcre = ActiveState::PCRE->new($filename);

Load the ActiveState::DB database from C<$filename>.

=item match()

 $bool = $pcre->match($id, $string);
 $bool = $pcre->match($id, $string, $offset);

Looks up the regular expression indexed by C<$id> and dies if the regular
expression does not exist. Returns 1 if the C<$string> matched or 0 otherwise.

C<$offset> defaults to C<0> if not specified.

=item match_asdb()

 $pcre->match_asdb($cb, $data, $asdb, $when, $save);
 See match_asdb_v2.

=item match_asdb_v2()

 $pcre->match_asdb_v2(\%args, @strings);

Where %args looks like:
 (
     asdb => hash of (regexp => label) tuples
     code => callback - see below
     data => data to match against
     when => 1 to call cb on a match; 0 to cb when there are no matches
     save => save the match?
 );

Enumerates through C<$asdb>; for each key/value pair, the key is assumed to
be the id of a compiled regexp, and is passed to C<pcre>'s C<match()> method.
If C<$save> is true, the value (the string that matched) is passed to the
callback. The callback is invoked each time a regular expression matches a
string. The callback should be written like this:

  sub cb {
      my ($data, $value, $matched) = @_;
  }

Where C<$data> was passed to C<match_asdb()>, C<$value> was the value in the
database that corresponds to the key used to look up the regular expression,
and C<$string> is the portion of the string that matched the regular
expression.

The callback should return 1 to continue scanning after the first match, or 0
to abort after the first successful match.

=back

In addition, ActiveState::PCRE optionally exports the following global
variables:

=over

=item @START

=item @END

These correspond to Perl's C<@-> and C<@+> offset arrays, respectively. That
is, the offset of the start and end of the last successful match.

=item $LIMIT

Set to 1 if pcre's internal recursion limit is exceeded, and remains at 1
until it is cleared.

=back

=head1 SEE ALSO

See also L<pcre>.

See also L<ActiveState::DB>.

=cut
