package ActiveState::RangeIndex;

use strict;
use Data::Dump 'dump';
use XSLoader ();
use base 'Exporter';

our $VERSION = '1.000';
XSLoader::load(__PACKAGE__, $VERSION);

our @EXPORT_OK = qw(ridx_lookup);

sub new {
    my $cls = shift;
    bless {
	packed => undef,
	ranges => []
    }, ref($cls) || $cls;
}

sub addrange {
    my ($self, $min, $max, $data) = @_;
    for ($min, $max, $data) {
	die "Can't handle non-integer data" unless /^\d+$/;
	die "Can't handle negative integers" if $_ < 0;
    }
    die "Bad range: max $max < min $min" if $max < $min;
    delete $self->{packed};
    push @{$self->{ranges}}, [$min, $max+1, $data];
}

sub pack {
    my $self = shift;
    # ranges are [min, max]

    my $points_of_light;
    for (@{$self->{ranges}}) {
	$points_of_light->{$_->[0]} = 1;
	$points_of_light->{$_->[1]} = 1;
    }
    my $steps;
    @$steps = sort {$a <=> $b} keys %$points_of_light;
    my $ranges = [];
    for my $i (0 .. $#$steps - 1) {
	my $start = $steps->[$i];
	my $end = $steps->[$i+1];
	my $data = [];
	for (@{$self->{ranges}}) {
	    my ($s,$e) = @$_;
	    if (
		($start >= $s && $start < $e) ||
		($end > $s    && $end < $e) ||
		($s >= $start && $s < $end) ||
		($e > $start  && $e < $end)
	    ) {
		push @$data, $_->[2];
	    }
	}
	if (@$data) {
	    push @$ranges, [$start, $end, $data];
	}
    }
    return $self->{packed} = ridx_pack($ranges);
}

sub query {
    my ($self, $id) = @_;
    $self->pack unless $self->{packed};
    return ridx_lookup($self->{packed}, $id);
}

sub maxvalues {
    my $self = shift;
    $self->pack unless $self->{packed};
    return ridx_maxvalues($self->{packed});
}

package ActiveState::RangeIndex::DB;

sub new {
    my ($class, $db, $key) = @_;
    my $self = bless [$db, $key, undef, undef], ref($class) || $class;
    # Get and cache the pointer and length of the underlying value.
    my ($ptr, $len) = $db->_get($key)
	or die "RangeIndex::DB: refresh: no such key $key";
    @$self[2, 3] = ($ptr, $len);
    $self;
}

sub destroy {
    my $self = shift;
    @$self = ();
}

sub query {
    my ($self, $id) = @_;
    return ActiveState::RangeIndex::ridx_lookup_ptr(@$self[2,3], $id);
}

sub maxvalues {
    my $self = shift;
    return ActiveState::RangeIndex::ridx_maxvalues_ptr(@$self[2,3]);
}

1;

=head1 NAME

ActiveState::RangeIndex - quickly find what ranges an integer is in

=head1 SYNPOSIS

 # Build a new range index
 $r = ActiveState::RangeIndex->new;
 $r->addrange(@$_) for (
#   START   END     VALUE (must be an integer)
    [0,     10,     42],
    [1000,  1040,   64],
    [5,     2000,   11],
 );

 # Look stuff up
 @matches = $r->query(8);      # returns (42, 11)
 @matches = $r->query(1000);   # returns (64, 11)

=head1 DESCRIPTION

ActiveState::RangeIndex associates an integer value with a range of integers.
When you query for a given integer C<I>, ActiveState::RangeIndex returns the
values whose ranges C<I> falls within.

For example, if you associate C<42> with range C<0 .. 10>, and C<24> with
range C<5 .. 15>, then querying for the numbers between C<0> and C<16> returns
the following results:

 QUERY    RESULTS
 0        42
 1        42
 2        42
 3        42
 4        42
 5        42 24
 6        42 24
 7        42 24
 8        42 24
 9        42 24
 10       42 24
 11       24
 12       24
 13       24
 14       24
 15       24
 16

The order of the results is not guaranteed. Everything must be an unsigned
integer, including the start and end offsets, the query terms, and the values
associated with each range.

=head2 ActiveState::RangeIndex API

ActiveState::RangeIndex provides the following methods:

=over

=item new()

 $ridx = ActiveState::RangeIndex->new;

Creates a new, empty RangeIndex object. Ranges can be added by calling the
C<addrange()> method.

=item addrange()

 $ridx->addrange($start, $end, $value);

Associates C<$value> with the range C<[$start .. $end]>, inclusive.

=item pack()

 $ridx->pack;

Compiles the range into a packed structure that is efficient to search
through. Returns the packed structure as a binary string.

=item query()

 @v = $ridx->query($index);

Looks up which ranges C<$index> falls into, and returns the corresponding
values.

Calls C<pack()> if it has not already been called.

=item maxvalues()

 $max = $ridx->maxvalues;

Returns the maximum number of values in any range, after all overlaps have
been taken into account.

=back

=head2 ActiveState::RangeIndex::DB

This is intended to support storing the packed ActiveState::RangeIndex in an
ActiveState::DB, without copying stuff.

Use it like this:

 $dbm = ActiveState::DB::Manager->new($db);
 $ridx = $dbm->nest_object("key", "ActiveState::RangeIndex::DB");
 @v = $ridx->query($ix);

Etc.

=head1 SEE ALSO

See also L<ActiveState::DB>.

=cut
