package ActiveState::Transformer::Plugin::GIFMetadata;

use strict;
use warnings;

use base qw(ActiveState::Transformer::Plugin);

# Transforms a GIF data stream into textual metadata, most values of which
# come directly from the various GIF headers, but also contains an expression
# of total bytes of content
#
# See http://www.w3.org/Graphics/GIF/spec-gif89a.txt


# state data -- reset after every call
our $pos;
our $imageref;
our $total_bytes;
our $total_pixels;
our %data;
our %orig_data;
our $eof;

use constant METADATA_FIELDS => qw(
    version
    width
    height
    colour_count
    colour_resolution
    gct_is_sorted
    bg_colour_index
    pixel_aspect_ratio

    disposal_method
    user_input_required
    has_transparency
    transparency_index
    delay_time

    x_offset
    y_offset
    image_width
    image_height
    image_colour_count
    is_interlaced
    lct_is_sorted

    image_aspect_ratio
    pixels_per_byte
);

use constant IMAGE                              => 0x2c;
use constant EXTENSION                          => 0x21;
use constant TERMINATOR                         => 0x3b;
use constant GRAPHICS_CONTROL_EXTENSION_BLOCK   => 0xF9;
use constant PLAIN_TEXT_EXTENSION_BLOCK         => 0x01;
use constant APPLICATION_EXTENSION_BLOCK        => 0xFF;
use constant COMMENT_EXTENSION_BLOCK            => 0xFE;

# block size byte plus the rest of the header
use constant PLAIN_TEXT_EXTENSION_BLOCK_LENGTH  => 1 + 0x0C;
use constant APPLICATION_EXTENSION_BLOCK_LENGTH => 1 + 0x0B;

# no header fields - data block starts right away
use constant COMMENT_EXTENSION_BLOCK_LENGTH     => 0;


# Runtime header specifications -- calculated once at compile time
our $cached_specs;

BEGIN {

    # This data is intended to go out of scope after the compile phase -- it
    # is transformed into runtime-usable data by the code below

    # structure: field name, pack specification, subroutine transformation (optional)
    my $headers = {
        header => { size => 6,
                    fields => [ [ 'version', 'A6', ],
            ],
        },
        logical_screen_descriptor => { size => 7,
                                       fields => [
                [ 'width',                  'v',    sub { $_[0] == 1 ? 1 : _quantize($_[0], 8) }, ],
                [ 'height',                 'v',    sub { $_[0] == 1 ? 1 : _quantize($_[0], 8) }, ],
                [ 'PACKED_LSD',             'C', ],
                [ 'bg_colour_index',        'C',    sub { $orig_data{colour_table_exists} ? $_[0] : 'undef' }, ],
                [ 'pixel_aspect_ratio',     'C',    sub { $_[0] > 0 ? int(($_[0] + 15) / 64) : 1 }, ],
            ],
        },
        image_block => { size => 9,
                         fields => [
                [ 'x_offset',               'v',    sub { _quantize($_[0], 8) }, ],
                [ 'y_offset',               'v',    sub { _quantize($_[0], 8) }, ],
                [ 'image_width',            'v',    sub { $_[0] == 1 ? 1 : _quantize($_[0], 8) }, ],
                [ 'image_height',           'v',    sub { $_[0] == 1 ? 1 : _quantize($_[0], 8) }, ],
                [ 'PACKED_IMAGE',           'C', ],
            ],
        },
        graphics_control_extension_block => { size => 6,
                                              fields => [
                [ 'gce_block_size',         'C', ],     # always 4
                [ 'PACKED_GCE',             'C', ],
                [ 'delay_time',             'v', ],
                [ 'transparency_index',     'C', sub { $orig_data{has_transparency} ? $_[0] : 'undef' }, ],
                [ 'gce_terminator',         'C', ],     # always 0 (zero-length data block)
            ],
        },
    };

    # structure: field name, offset, length, subroutine transformation (optional)
    my $bitspec = {
        PACKED_LSD => [
            [ 'colour_table_exists',    0, 1, ],
            [ 'colour_resolution',      1, 3, sub { $_[0] + 1 }, ],
            [ 'gct_is_sorted',          4, 1, sub { $orig_data{colour_table_exists} ? _boolean($_[0]) : 'undef' }, ],
            [ 'colour_count',           5, 3, sub { 2 ** ($_[0] + 1) }, ],
        ],
        PACKED_IMAGE => [
            [ 'local_colour_table_exists', 0, 1 ],
            [ 'is_interlaced',          1, 1, \&_boolean, ],
            [ 'lct_is_sorted',          2, 1, sub { $orig_data{local_colour_table_exists} ? _boolean($_[0]) : 'undef' }, ],
            # [ 'reserved', 3, 2, ],
            [ 'image_colour_count',     5, 3, sub { $orig_data{local_colour_table_exists} ? 2 ** ($_[0] + 1) : 0 }, ],
        ],
        PACKED_GCE => [
            # [ 'reserved', 0, 3, ],
            [ 'disposal_method',        3, 3, sub { qw(unspecified do_not_dispose restore_to_background restore_to_previous)[$_[0]] || 'unknown' }, ],
            [ 'user_input_required',    6, 1, \&_boolean, ],
            [ 'has_transparency',       7, 1, \&_boolean, ],
        ],
    };

    # Note that all this is happening at compile-time -- we take the more
    # human-readable and -editable data above and transform it into a form
    # that is faster to parse at runtime
    my %subs;
    foreach my $block (keys %$headers) {
        $cached_specs->{$block} = {
            packed => join('', map { $_->[1] } @{$headers->{$block}{fields}}),
            fields => [ map { $_->[0] } @{$headers->{$block}{fields}} ],
            size => $headers->{$block}{size},
        };
        $cached_specs->{$block}{bitfield} = ( grep { $_ =~ m/^PACKED_/ } @{$cached_specs->{$block}{fields}} )[0];

        my %temp_sub = map { $_->[2] ? ($_->[0] => $_->[2]) : () } @{$headers->{$block}{fields}};
        @subs{keys %temp_sub} = values %temp_sub;
    }

    foreach my $bitfield (keys %$bitspec) {
        $cached_specs->{$bitfield} = {
            fields => [ map { $_->[0] } @{$bitspec->{$bitfield}} ],
            offsets => { map { $_->[0] => [ $_->[1], $_->[2] ] } @{$bitspec->{$bitfield}} },
        };
        my %temp_sub = map { $_->[3] ? ($_->[0] => $_->[3]) : () } @{$bitspec->{$bitfield}};
        @subs{keys %temp_sub} = values %temp_sub;
    }

    $cached_specs->{subs} = \%subs;

}; # end BEGIN


sub transform {
    my $self = shift;

    $pos = 0;
    $imageref = \$_[0];
    $total_bytes = 0;
    $total_pixels = 0;
    %data = ();
    %orig_data = ();
    $eof = 0;

    _get_content('header');
    return '' if $data{version} !~ m/GIF8[79]a/;

    _get_content('logical_screen_descriptor');

    if ($orig_data{colour_table_exists}) {
        _seek_cur(3 * $data{colour_count});
    }

    my ($seen_image_block, $seen_gce) = (0, 0);
    while (not $eof and (not $seen_image_block or not $seen_gce)) {
        my $block_descriptor = _read(1, 'C');
        last if $eof or not defined $block_descriptor;

        if ($block_descriptor == TERMINATOR) {
            last;
        }

        elsif ($block_descriptor == IMAGE) {
            last if $seen_image_block;
            $seen_image_block = 1;

            _get_content('image_block');
            $total_pixels += $orig_data{image_width} * $orig_data{image_height}
                if $orig_data{image_width} and $orig_data{image_height};
            _seek_cur(3 * $data{image_colour_count})
                if $orig_data{local_colour_table_exists};

            # LZW header and data
            _seek_cur(1);
            $total_bytes += _read_data_blocks();
            next;
        }

        elsif ($block_descriptor == EXTENSION) {
            # these can only be *before* their corresponding image
            last if $seen_gce or $seen_image_block;

            my $extension_type = _read(1, 'C');
            last if $eof;
            if ($extension_type == GRAPHICS_CONTROL_EXTENSION_BLOCK) {
                $seen_gce = 1;
                _get_content('graphics_control_extension_block');
            }
            elsif ($extension_type == PLAIN_TEXT_EXTENSION_BLOCK) {
                _seek_cur(PLAIN_TEXT_EXTENSION_BLOCK_LENGTH);
                _read_data_blocks();
            }
            elsif ($extension_type == APPLICATION_EXTENSION_BLOCK) {
                _seek_cur(APPLICATION_EXTENSION_BLOCK_LENGTH);
                _read_data_blocks();
            }
            elsif ($extension_type == COMMENT_EXTENSION_BLOCK) {
                _seek_cur(COMMENT_EXTENSION_BLOCK_LENGTH);
                _read_data_blocks();
            }
            else {
                # unknown extension type
                last;
            }
            next;
        }

        else {
            # unknown block type
            last;
        }
    }

    $data{image_aspect_ratio} = _truncate($orig_data{image_width} / $orig_data{image_height})
        if $orig_data{image_width} and $orig_data{image_height};

    # The caller can override the calculated total_bytes
    $total_bytes = $_[1] if defined $_[1];

    $data{pixels_per_byte} = ((not $total_bytes) ? 'unknown' :
        (_quantize(int($total_pixels / $total_bytes), 2) + 1));

    # If none of the METADATA_FIELDS are present in the data, then we're not
    # going to want to include the mime_type too.
    my $result = join(' ', map { ( defined $data{$_} ? $_ . '=' . $data{$_} : ()) } METADATA_FIELDS);
    $result .= ' mime_type=image/gif' if $result;
    return $result;
}


# offset and length are like substr, counting from the beginning of the byte
# convert to unsigned char FIRST, if necessary, via ord() or unpack('C')
sub _get_bits {
    # my $content = shift;
    # my $offset = shift;
    # my $length = shift;

    # use the length to compute the mask
    my $mask = (1 << $_[2]) - 1;

    # and shift it to the right place
    my $shift = 8 - $_[1] - $_[2];
    $mask = $mask << $shift;

    return ($_[0] & $mask) >> $shift;
}

sub _boolean {
    return $_[0] ? 'true' : 'false';
}

sub _quantize {
    # my ($val, $amt) = @_;
    return $_[0] - ($_[0] % $_[1]);
}

# truncates to one decimal place and always expresses as fixed point
sub _truncate {
    my $val = int($_[0] * 10);
    substr($val, -1, 0) = '.';
    return $val;
}

# arg0: length in bytes
# arg1: unpack specification (optional)
sub _read {
    my $size = shift;
    my $packspec = shift;
    no warnings 'substr';
    my $content = substr($$imageref, $pos, $size);
    $pos += $size;
    my $length;
    if (not defined $content or not ($length = length $content)) {
        $eof = 1;
        return '';
    }

    $eof = 1 if $length != $size;
    return $content if not $packspec;
    return unpack($packspec, $content);
}

sub _seek_cur {
    $pos += $_[0];
}

# Note that "bytes_read" may overrun the end of the data provided, since we
# don't actually read it, just use the length indicator as a guide
sub _read_data_blocks {
    my $bytes_read = 0;
    while (1) {
        my $block_length = _read(1, 'C');
        last if not $block_length;
        $pos += $block_length;
        $bytes_read += $block_length;
    }
    return $bytes_read;
}

# arg: name of block of data we care about -- index into specifications hash
# We also post-process the data here, via the subrefs provided
# return value not used yet -- but indicates success/failure
sub _get_content {
    my $block = shift;

    my $packspec = $cached_specs->{$block}{packed};
    my @fields = @{$cached_specs->{$block}{fields}};
    my $size = $cached_specs->{$block}{size};
    my $content = _read($size);

    # quit if we failed to read anything, otherwise soldier on and make
    # the best of a bad situation...
    return 0 if $eof and not defined $content;

    @orig_data{@fields} = unpack($packspec, $content);

    my $bitfield = $cached_specs->{$block}{bitfield};
    if ($bitfield and defined $orig_data{$bitfield}) {
        my %temp_data = map
            { $_ => _get_bits($orig_data{$bitfield},
                    @{$cached_specs->{$bitfield}{offsets}{$_}}) }
            @{$cached_specs->{$bitfield}{fields}};
        @orig_data{keys %temp_data} = values %temp_data;
        push @fields, @{$cached_specs->{$bitfield}{fields}};
    }

    # post-process the data as expected
    map { $data{$_} = ( (exists $cached_specs->{subs}{$_} and defined $orig_data{$_}) ?
                        $cached_specs->{subs}{$_}->($orig_data{$_}) :
                        $orig_data{$_}
                      ) } @fields;

    return 1;
}


1;
