package ActiveState::Transformer::Plugin::HTMLSimplify;
use strict;
use warnings;
use base qw(ActiveState::Transformer::Plugin);


our $FULL_TAG = qr{<[^>]+>};

our $TAG_NAME = qr{[^'"/> ]+};

our $NULLIFY_TAGS = qr{(?=[pstx])(?:plaintext|script|textarea|xmp)}i;

our $URI_SCHEMES = qr{(?=[cdfhm])(?:(?=[cdm])(?:cid|data|mailto|mid):|(?=[fh])(?:ftp|https?)://)}i;

our $HEX_COLOUR = qr{#[0-9a-f]{3}(?:[0-9a-f]{3})?}i;

# Hash of html tags that are especially likely to contain uris, and the
# attribute where they can be found -- anything in these attributes will
# automatically be normalized (even if they don't match what we would
# normally call a uri) -- and moreover we don't do any further searching
# inside html tags other than in the attributes we list here.

our %uri_tag_attr = (
    ( map { $_ => 'href' }       qw(a area link) ),
    ( map { $_ => 'src' }        qw(img frame iframe embed ) ),
    ( map { $_ => 'background' } qw(body table tr td) ),
    ( map { $_ => 'action' }     qw(form) ),
    ( map { $_ => 'codebase' }   qw(object) ),
);


sub transform {
    my $self = shift;
    my $text = shift;

    # strip everything not contained in the outermost 'tag'
    $text =~ s/^[^<]+//;
    $text =~ s/[^>]+$//;

    # collapse whitespace
    $text =~ s/\s+/ /g;

    # remove CDATA
    $text =~ s/<!\[CDATA\[.*?\]\]>//gi;

    # remove comments
    $text =~ s/<!--.+?--( ?)>/<!-- -->/g;

    my $result = '';            # FYI it is *not* faster to preallocate the string
    my $nullify_depth = 0;      # how many nullifying tags deep we are

    while ($text =~ m/\G.*?($FULL_TAG)/go) {
        my $full_tag = $1;

        my $tag_name = ($full_tag =~ m/<\/?\s?($TAG_NAME)/o)[0];

        if (substr($full_tag, 1, 1) ne '/') {
            $result .= _transform_tag($tag_name, $full_tag) unless $nullify_depth;
            $nullify_depth += 1 if defined $tag_name and $tag_name =~ m/^${NULLIFY_TAGS}$/o;
        }
        else {
            $nullify_depth -= 1 if $nullify_depth and defined $tag_name and $tag_name =~ m/^$NULLIFY_TAGS$/o;
            $result .= _transform_tag($tag_name, $full_tag) unless $nullify_depth;
        }

        # Now handle the text after the tag (if any), or just skip to the next
        # tag if we are eating all the content
        next if $nullify_depth;

        # Quantify _all_ content between <style>foo</style> tags
        if ($tag_name && substr($full_tag, 1, 1) ne '/' && $tag_name =~ m/^(?:style)$/) {
            if ($text =~ /\G(.*?)(<\/\s*$tag_name\s*>)/gc) {
                $result .= _quantify_text($1) . $2;
            }
        }

        if ($text =~ /\G([^<]+)/gc) {
            $result .= _quantify_text($1);
        }
    }

    return $result;
}

sub _transform_tag {
    my $tag_name = shift;
    my $full_tag = shift;

    return $full_tag if $full_tag eq '<!-- -->';

    # This is coded carefully so as to allow future extraction of content
    # so as to obviate the use of HTML::Parser in the part plugin later

    my $lc_tag_name = lc $tag_name;

    # remove img alt values
    if ($lc_tag_name eq 'img') {
        $full_tag =~ s/(?<=alt)(?:(="[^"]*")|(='[^']*')|(=[^> ]+))//gi;
        pos($full_tag) = 0;
        # List of attributes which will be replaced by quantified text
        for my $attr (qw/title/) {
            while ($full_tag =~ m/\G.*?$attr=/gi) {
                my $c2 = pos($full_tag);
                $full_tag =~ s/\G(['"])(.*?)(?=['"][ >])/$1 . _quantify_text($2) /e;
                pos($full_tag) = $c2;
                $full_tag =~ s/\G([^'"].*?)(?=[ >])/ _quantify_text($1) /e;
                pos($full_tag) = $c2;
            }
        }
    }
    if ($lc_tag_name eq 'meta') {
        $full_tag =~ s/(?<=content)(?:(="[^"]*")|(='[^']*')|(=[^> ]+))//gi;
    }

    # normalize uris - inside certain tags only
    if (my $uri_attr = $uri_tag_attr{$lc_tag_name}) {
        while ($full_tag =~ m/\G.*?$uri_attr=/gi) {
            my $cached_pos = pos($full_tag);
            $full_tag =~ s/\G(['"]*(?:$URI_SCHEMES)?).+?(?=['"]*[ >])/$1~URI~/o;
            pos($full_tag) = $cached_pos;
        }
    }

    # normalize colours
    while ($full_tag =~ m/(?<=['"= ])($HEX_COLOUR)(?=['"]*[ >])/o) {
        my $hex_colour = $1;
        my $new_colour = _transform_colour($hex_colour);
        $full_tag =~ s/$hex_colour/$new_colour/;
    }

    return $full_tag;
}

sub _transform_colour {
    my $text = shift;

    return $text if substr($text, 0, 1) ne '#';

    my $hex_colour = substr($text, 1);
    my $length = length($hex_colour);

    my ($red, $green, $blue);

    if ($length == 3) {
        ($red, $green, $blue) = map {$_ .= $_} split('', $hex_colour);
    }
    elsif ($length == 6) {
        ($red, $green, $blue) = ($hex_colour =~ /../g);
    }
    else {
        return $text;
    }

    ($red, $green, $blue) = map { int((hex($_) + 1) / 32) } ($red, $green, $blue);

    return "#~$red$green$blue~";
}

sub _quantify_text {
    my $text = shift;

    return '' unless $text;

    # We pretend a blank space is of zero length, but keep the space intact.
    return $text if $text eq ' ';

    my $length = length $text;

    return '~CHARACTER~'    if $length == 1;
    return '~VERY_LITTLE~'  if $length <= 5;
    return '~LITTLE~'       if $length <= 49;
    return '~MEDIUM~'       if $length <= 99;
    return '~MUCH~'         if $length <= 249;
    return '~VERY_MUCH~';
}

1;
