package ActiveState::Transformer::Plugin::JPEGMetadata;

use strict;
use warnings;

use base qw(ActiveState::Transformer::Plugin);

# Transforms a JPEG data stream into textual metadata, most values of which
# come directly from the various JPEG headers, but also contains an expression
# of total bytes of content
#
# See http://FIXME...

my %exif_to_spops = (
    ImageWidth => 'width',
    ImageHeight => 'height',
    XResolution => 'XResolution',
    YResolution => 'YResolution',
    'thumb_jfif_x' => 'thumb_jfif_x',
    'thumb_jfif_y' => 'thumb_jfif_y',
    ExifThumbnailImageWidth => 'exif_thumbnail_width',
    ExifThumbnailImageHeight => 'exif_thumbnail_height',
);

my %stream_types = (
    0xc0 => 'nondifferential_huffman_baseline',
    0xc1 => 'nondifferential_huffman_sequential',
    0xc2 => 'nondifferential_huffman_progressive',
    0xc3 => 'nondifferential_huffman_lossless',
    0xc5 => 'differential_huffman_sequential',
    0xc6 => 'differential_huffman_progressive',
    0xc7 => 'differential_huffman_lossless',
    0xc9 => 'nondifferential_arithmetic_sequential',
    0xca => 'nondifferential_arithmetic_progressive',
    0xcb => 'nondifferential_arithmetic_lossless',
    0xcd => 'differential_arithmetic_sequential',
    0xce => 'differential_arithmetic_progressive',
    0xcf => 'differential_arithmetic_lossless',
);

my %ignore_if_zero = (
    thumb_jfif_x => 1,
    thumb_jfif_y => 1,
);

my %normalization_table = (
    stream_type => undef,
    width => 16,
    height => 16,
    segments => undef,
    mime_type => undef,
    pixels_per_byte => 1,
);


sub new {
    my ($class, $this) = @_;
    $class = 'ActiveState::Transformer::Plugin::JPEGMetadata' unless $class;
    $this = {} unless $this;

    my $self = $class->SUPER::new($this);
    $self->{exif_tool} = $self->get_exiftool();
    $self->{normalization_table} = \%normalization_table unless $self->{normalization_table};
    return $self;
}

# Obtains metadata from a binary JPEG image
# @param self    transformer object
# @param imageref    binary JPEG image
# @param real_length    original length of the JPEG image
sub transform {
    my $self = shift;
    my $imageref = \$_[0];
    my $real_length = $_[1];

    my $options = {
        Sort => 'File',
        ReportExifSegments => 1,
        ReportXMPSegments => 1,
        ReportIPTCSegments => 1,
        ReportSOF => 1,
        ReportSegments => 1,
        ReportSegmentsOnly => 1,
        TagTableWarnings => 0,
        FastScan => 1,
        IgnoreTrailers => 1,
    };

    # Parse the jpeg image using ExifTool
    my $info = $self->{exif_tool}->ImageInfo($imageref, $options);

    my @metadata = ();
    my @segments = ();
    my $thumbnail_index = undef;

    # GetFoundTags gives us the tags in the order they are found in the file
    for my $tag ($self->{exif_tool}->GetFoundTags('File')) {
        if (exists $exif_to_spops{$tag}) {
            unless (defined $ignore_if_zero{$tag} && $info->{$tag} eq '0') {
                push @metadata, "$exif_to_spops{$tag}=$info->{$tag}" if defined $info->{$tag};
            }
        } elsif ($tag =~ m/^exif_(.*)$/) {
            push @metadata, "exif_" . $self->url_encode($1) . "=" . $self->url_encode($info->{$tag});
        } elsif ($tag eq 'has_iptc') {
            push @metadata, 'has_IPTC=1' if $info->{$tag};
        } elsif ($tag eq 'has_xmp') {
            push @metadata, 'has_XMP=1' if $info->{$tag};
        } elsif ($tag eq 'has_exif') {
            push @metadata, 'has_Exif=1' if $info->{$tag};
        } elsif ($tag eq 'SOF') {
            if (exists $stream_types{$info->{$tag}}) {
                push @metadata, "stream_type=$stream_types{$info->{$tag}}";
            }
        } elsif ($tag eq 'ThumbnailImage') {
            $thumbnail_index = scalar @metadata;
            push @metadata, '', '';
        } elsif ($tag eq 'Comment') {
            push @metadata, "comment=" . $self->url_encode(substr($info->{$tag}, 0, 100));
        } elsif ($tag =~ m/^SEGMENT_(.*)/)
        {
            my $type = $1;
            # remove count added by ExifTool intended to differenciate between duplicate tags
            $type =~ s/\s\(\d+\)$//;
            push @segments, $type;
        } elsif ($tag =~ m/^SEGMENTTYPE_(.*)/)
        {
            my $type = $1;
            $type =~ s/\s\(\d+\)$//;
            $segments[-1] .= "($type)" if @segments;
#        } else {
#            print STDERR "DEBUG: unknown tag '$tag' - data=$info->{$tag}\n";
        }
    }

    push @metadata, 'segments=' . join(",", @segments) if @segments;

    my ($aspect_ratio, $pixels_per_byte);
    if ($info->{ImageWidth} && $info->{ImageHeight}) {
        $aspect_ratio = sprintf("%.1f", $info->{ImageWidth} / $info->{ImageHeight});
        if ($real_length == 0) {
            $pixels_per_byte = '0.0';
        }
        else {
            $pixels_per_byte = sprintf("%.1f", ($info->{ImageWidth} * $info->{ImageHeight}) / $real_length);
        }
    }
    else {
        $aspect_ratio = '0.0';
        $pixels_per_byte = '0.0';
    }

    push @metadata, "aspect_ratio=$aspect_ratio";
    push @metadata, "pixels_per_byte=$pixels_per_byte";

    # Always include the actual mime type
    push @metadata, 'mime_type=image/jpeg';


    if (defined $thumbnail_index) {
        my @taglist = qw/ImageHeight ImageWidth/;
        my $thumbnail_options = {
            Sort => 'File',
        };

        my $thumb_info = $self->{exif_tool}->ImageInfo($info->{ThumbnailImage}, \@taglist, $thumbnail_options);

        for my $tag ($self->{exif_tool}->GetFoundTags('File')) {
            if ($tag eq 'ImageHeight' || $tag eq 'ImageWidth') {
                $metadata[$thumbnail_index++] = $exif_to_spops{"ExifThumbnail$tag"} . "=$thumb_info->{$tag}";
            } 
        }
    }
#    print STDERR "DEBUG: " . join(' ', @metadata) . "\n";

    return join(' ', @metadata);
}

1;
