package ActiveState::Transformer::Plugin::PNGMetadata;

use strict;
use warnings;

use base qw(ActiveState::Transformer::Plugin);

use Time::Local qw(timegm);
# Transforms a PNG data stream into textual metadata, most values of which
# come directly from the various PNG headers, but also contains an expression
# of total bytes of content
#
# See http://www.w3.org/Graphics/PNG
# man Image::ExifTool::TagNames for more info on tag names

my %exif_to_spops = (
    ImageWidth => 'width',
    ImageHeight => 'height',
    PixelsPerUnitX => 'phys_x_ppu',
    PixelsPerUnitY => 'phys_y_ppu',
    PixelUnits => 'phys_unit',
);

# Example of metadata from a PNG file:
# width=20 height=10 aspect_ratio=2.0 pixels_per_byte=0.5 chunks=IHDR,bKGD,vpAg,IDAT,IEND mime_type=image/png
# in this table:
#   - a number means normalize by rounding down to the nearest multiple of that number
#   - undef means pass that key value pair straight through if it is present
#   - a ref to an array means discretize to the even elements using the odd elements as boundries
# see unit tests for exact functionality
my %normalization_table = (
    width => 16,
    height => 16,
    pixels_per_byte => 1,
    chunks => undef,
    mime_type => undef,
);
    
sub new {
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{exif_tool} = $self->get_exiftool();
    $self->{normalization_table} = \%normalization_table unless $self->{normalization_table};
    $self->{no_extra_tags} = 1;
    return $self;
}

sub transform {
    my $self = shift;
    my $imageref = \$_[0];
    my $real_length = $_[1];

    my $options = {
        Sort => 'File',
        PrintConv => 0,
        AllowSpacesInTags => 1,
        EncodeITXTlanguage => 1,
        AddTextTagPrefix => 1,
        TagTableWarnings => 0,
    };

    my $info = $self->{exif_tool}->ImageInfo($imageref, $options);
    
    my $metadata = '';

    for my $tag ($self->{exif_tool}->GetFoundTags('File')) {
        if (defined $info->{$tag}) {
            if (exists $exif_to_spops{$tag}) {
                $metadata .= "$exif_to_spops{$tag}=$info->{$tag} ";
            } elsif ($tag =~ m/^text_(.*$)/) {
                $metadata .= "text_" . $self->url_encode($1) . '=' . $self->url_encode($info->{$tag}) . ' ';
            } elsif ($tag =~ m/^itxt_(.*$)/) {
                my ($realtag, $lang, $trans) = split /\0/, $1;

                $metadata .= "itxt_" . $self->url_encode($realtag) . '/' . $self->url_encode($lang) . '/' . $self->url_encode($trans) . '=' . $self->url_encode($info->{$tag}) . ' ';
            } elsif ($tag eq 'ModifyDate') {
                my $date = $info->{$tag};
                $date =~ s/[\:| ]//g;
                $metadata .= "time=$date ";

                if($date =~ m/(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})/) {
                    my $age = eval { 
                        sprintf("%.1f", (time - timegm($6, $5, $4, $3, $2 - 1, $1)) / 86400.0);
                    };
                    if ($@) {
                        $metadata .= "age=invalid_date ";
                    } else {
                        $metadata .= "age=$age ";
                    }
                }
#           } else {
#               print STDERR "DEBUG: not recognised: $tag\n";
            }
        }
    }

    my ($aspect_ratio, $pixels_per_byte);
    if ($info->{ImageWidth} && $info->{ImageHeight}) {
        $aspect_ratio = sprintf("%.1f", $info->{ImageWidth} / $info->{ImageHeight});
        if ($real_length == 0) {
            $pixels_per_byte = '0.0';
        }
        else {
            $pixels_per_byte = sprintf("%.1f", ($info->{ImageWidth} * $info->{ImageHeight}) / $real_length);
        }
    }
    else {
        $aspect_ratio = '0.0';
        $pixels_per_byte = '0.0';
    }

    $metadata .= "aspect_ratio=$aspect_ratio ";
    $metadata .= "pixels_per_byte=$pixels_per_byte ";

    my @chunks;
    push(@chunks, $self->url_encode($_)) foreach @{$self->{exif_tool}->{'CHUNK_LIST'}};
    my $chunklist = 'chunks=' . join(',', @chunks);

    $metadata .= "$chunklist mime_type=image/png";

    return $metadata;
}

1;
